@@ -18,6 +18,10 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
 
     MVMint32 mro_idx = MVM_repr_elems(tc, mro);
 
+    MVM_gc_root_temp_push(tc, (MVMCollectable **)&st);
+    MVM_gc_root_temp_push(tc, (MVMCollectable **)&mro);
+
+    /* We're only ever called with tc->allocate_in_gen2 set, so no need to MVMROOT here. */
     flat_list = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
 
     class_list = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
@@ -37,8 +41,13 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
         if (num_parents <= 1) {
             /* Get attributes and iterate over them. */
             MVMObject *attributes     = MVM_repr_at_pos_o(tc, type_info, 1);
-            MVMIter * const attr_iter = (MVMIter *)MVM_iter(tc, attributes);
             MVMObject *attr_map = NULL;
+            MVMIter   *attr_iter;
+
+            /* Add to class list firstly, so we can avoid a MVMROOT. */
+            MVM_repr_push_o(tc, class_list, current_class);
+
+            attr_iter = (MVMIter *)MVM_iter(tc, attributes);
 
             if (MVM_iter_istrue(tc, attr_iter)) {
                 attr_map = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_hash_type);
@@ -66,8 +75,7 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
                 MVM_repr_push_o(tc, flat_list, attr);
             }
 
-            /* Add to class list and map list. */
-            MVM_repr_push_o(tc, class_list, current_class);
+            /* Add to map list. */
             MVM_repr_push_o(tc, attr_map_list, attr_map);
         }
         else {
@@ -77,6 +85,8 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
         }
     }
 
+    MVM_gc_root_temp_pop_n(tc, 2); /* mro, st */
+
     /* We can now form the name map. */
     num_classes = MVM_repr_elems(tc, class_list);
     result = (MVMCPPStructNameMap *) MVM_malloc(sizeof(MVMCPPStructNameMap) * (1 + num_classes));
@@ -107,9 +117,9 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                                         MVMObject *repr_info, MVMCPPStructREPRData *repr_data) {
     /* Compute index mapping table and get flat list of attributes. */
     MVMObject *flat_list;
-    MVMROOT(tc, st, {
+    MVMROOT(tc, st) {
         flat_list = index_mapping_and_flat_list(tc, repr_info, repr_data, st);
-    });
+    }
 
     /* If we have no attributes in the index mapping, then just the header. */
     if (repr_data->name_to_index_mapping[0].class_key == NULL) {
@@ -166,6 +176,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                 MVMint32  type_id    = REPR(type)->ID;
                 if (spec->inlineable == MVM_STORAGE_SPEC_INLINED &&
                         (spec->boxed_primitive == MVM_STORAGE_SPEC_BP_INT ||
+                         spec->boxed_primitive == MVM_STORAGE_SPEC_BP_UINT64 ||
                          spec->boxed_primitive == MVM_STORAGE_SPEC_BP_NUM)) {
                     /* It's a boxed int or num; pretty easy. It'll just live in the
                      * body of the struct. Instead of masking in i here (which
@@ -377,11 +388,11 @@ static MVMint32 try_get_slot(MVMThreadContext *tc, MVMCPPStructREPRData *repr_da
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
     MVMSTable *st  = MVM_gc_allocate_stable(tc, &CPPStruct_this_repr, HOW);
 
-    MVMROOT(tc, st, {
+    MVMROOT(tc, st) {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
         MVM_ASSIGN_REF(tc, &(st->header), st->WHAT, obj);
         st->size = sizeof(MVMCPPStruct);
-    });
+    }
 
     return st->WHAT;
 }
@@ -403,7 +414,7 @@ static void initialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, voi
 
     /* Allocate object body. */
     MVMCPPStructBody *body = (MVMCPPStructBody *)data;
-    body->cppstruct = MVM_calloc(1, repr_data->struct_size > 0 ? repr_data->struct_size : 1);
+    body->cppstruct = calloc(1, repr_data->struct_size > 0 ? repr_data->struct_size : 1);
 
     /* Allocate child obj array. */
     if (repr_data->num_child_objs > 0)
@@ -415,7 +426,7 @@ static void initialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, voi
         MVMint32 i;
         for (i = 0; repr_data->initialize_slots[i] >= 0; i++) {
             MVMint32  offset = repr_data->struct_offsets[repr_data->initialize_slots[i]];
-            MVMSTable *st     = repr_data->flattened_stables[repr_data->initialize_slots[i]];
+            MVMSTable *st    = repr_data->flattened_stables[repr_data->initialize_slots[i]];
             st->REPR->initialize(tc, st, root, (char *)body->cppstruct + offset);
         }
     }
@@ -471,7 +482,7 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                     /* No cached object. */
                     void *cobj = get_ptr_at_offset(body->cppstruct, repr_data->struct_offsets[slot]);
                     if (cobj) {
-                        MVMROOT(tc, root, {
+                        MVMROOT(tc, root) {
                             if (type == MVM_CPPSTRUCT_ATTR_CARRAY) {
                                 obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
                             }
@@ -500,13 +511,13 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                                 obj = MVM_nativecall_make_cpointer(tc, typeobj, cobj);
                             }
                             else if(type == MVM_CPPSTRUCT_ATTR_STRING) {
-                                MVMROOT(tc, typeobj, {
+                                MVMROOT(tc, typeobj) {
                                     MVMString *str = MVM_string_utf8_decode(tc, tc->instance->VMString,
                                         cobj, strlen(cobj));
                                     obj = MVM_repr_box_str(tc, typeobj, str);
-                                });
+                                }
                             }
-                        });
+                        }
                         MVM_ASSIGN_REF(tc, &(root->header), body->child_objs[real_slot], obj);
                     }
                     else {
@@ -525,6 +536,14 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                 MVM_exception_throw_adhoc(tc, "CPPStruct: invalid native get of object attribute");
             break;
         }
+        case MVM_reg_uint64: {
+            if (attr_st)
+                result_reg->u64 = attr_st->REPR->box_funcs.get_uint(tc, attr_st, root,
+                    ((char *)body->cppstruct) + repr_data->struct_offsets[slot]);
+            else
+                MVM_exception_throw_adhoc(tc, "CPPStruct: invalid native get of object attribute");
+            break;
+        }
         case MVM_reg_num64: {
             if (attr_st)
                 result_reg->n64 = attr_st->REPR->box_funcs.get_num(tc, attr_st, root,
@@ -636,6 +655,14 @@ static void bind_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                 MVM_exception_throw_adhoc(tc, "CPPStruct: invalid native binding to object attribute");
             break;
         }
+        case MVM_reg_uint64: {
+            if (attr_st)
+                attr_st->REPR->box_funcs.set_uint(tc, attr_st, root,
+                    ((char *)body->cppstruct) + repr_data->struct_offsets[slot], value_reg.u64);
+            else
+                MVM_exception_throw_adhoc(tc, "CPPStruct: invalid native binding to object attribute");
+            break;
+        }
         case MVM_reg_num64: {
             if (attr_st)
                 attr_st->REPR->box_funcs.set_num(tc, attr_st, root,
@@ -740,7 +767,7 @@ static void gc_cleanup(MVMThreadContext *tc, MVMSTable *st, void *data) {
     /* XXX For some reason, this causes crashes at the moment. Need to
      * work out why. */
     /*if (body->cppstruct)
-        MVM_free(body->cppstruct);*/
+        free(body->cppstruct);*/
 }
 
 /* Called by the VM in order to free memory associated with this object. */
@@ -808,9 +835,7 @@ static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerial
     MVMint32 i, num_classes, num_slots;
 
     repr_data->struct_size    = MVM_serialization_read_int(tc, reader);
-    if (reader->root.version >= 17) {
-        repr_data->struct_align = MVM_serialization_read_int(tc, reader);
-    }
+    repr_data->struct_align   = MVM_serialization_read_int(tc, reader);
     repr_data->num_attributes = MVM_serialization_read_int(tc, reader);
     repr_data->num_child_objs = MVM_serialization_read_int(tc, reader);
 
