@@ -8,11 +8,11 @@ static const MVMREPROps StaticFrameSpesh_this_repr;
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
     MVMSTable *st  = MVM_gc_allocate_stable(tc, &StaticFrameSpesh_this_repr, HOW);
 
-    MVMROOT(tc, st, {
+    MVMROOT(tc, st) {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
         MVM_ASSIGN_REF(tc, &(st->header), st->WHAT, obj);
         st->size = sizeof(MVMStaticFrameSpesh);
-    });
+    }
 
     return st->WHAT;
 }
@@ -26,14 +26,13 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorklist *worklist) {
     MVMStaticFrameSpeshBody *body = (MVMStaticFrameSpeshBody *)data;
     MVM_spesh_stats_gc_mark(tc, body->spesh_stats, worklist);
-    MVM_spesh_arg_guard_gc_mark(tc, body->spesh_arg_guard, worklist);
+    MVM_spesh_arg_guard_gc_mark(tc, (MVMSpeshArgGuard *)MVM_load(&body->spesh_arg_guard), worklist);
     if (body->num_spesh_candidates) {
         MVMuint32 i;
         for (i = 0; i < body->num_spesh_candidates; i++) {
             MVM_gc_worklist_add(tc, worklist, &body->spesh_candidates[i]);
         }
     }
-    MVM_gc_worklist_add(tc, worklist, &body->plugin_state);
 }
 
 /* Called by the VM in order to free memory associated with this object. */
@@ -41,7 +40,8 @@ static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
     MVMStaticFrameSpesh *sfs = (MVMStaticFrameSpesh *)obj;
     MVM_spesh_stats_destroy(tc, sfs->body.spesh_stats);
     MVM_free(sfs->body.spesh_stats);
-    MVM_spesh_arg_guard_destroy(tc, sfs->body.spesh_arg_guard, 0);
+    MVM_spesh_arg_guard_destroy(tc, (MVMSpeshArgGuard *)MVM_load(&sfs->body.spesh_arg_guard), 0);
+    MVM_free(sfs->body.spesh_candidates);
 }
 
 static const MVMStorageSpec storage_spec = {
