@@ -5,6 +5,10 @@
 #define snprintf _snprintf
 #endif
 
+#ifndef AI_NUMERICSERV
+#define AI_NUMERICSERV 0
+#endif
+
 /* Assumed maximum packet size. If ever changing this to something beyond a
  * 16-bit number, then make sure to change the receive offsets in the data
  * structure below. */
@@ -185,7 +189,6 @@ static void socket_truncate(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 bytes
 /* Writes the specified bytes to the stream. */
 static MVMint64 socket_write_bytes(MVMThreadContext *tc, MVMOSHandle *h, char *buf, MVMuint64 bytes) {
     MVMIOSyncSocketData *data = (MVMIOSyncSocketData *)h->body.data;
-    MVMint64 sent = 0;
     unsigned int interval_id;
 
     interval_id = MVM_telemetry_interval_start(tc, "syncsocket.write_bytes");
@@ -200,7 +203,6 @@ static MVMint64 socket_write_bytes(MVMThreadContext *tc, MVMOSHandle *h, char *b
             MVM_telemetry_interval_stop(tc, interval_id, "syncsocket.write_bytes");
             throw_error(tc, r, "send data to socket");
         }
-        sent += r;
         buf += r;
         bytes -= r;
     }
@@ -509,7 +511,8 @@ static const MVMIOSockety             sockety = { socket_connect,
                                                   socket_accept,
                                                   socket_getport };
 static const MVMIOIntrospection introspection = { socket_is_tty,
-                                                  socket_handle };
+                                                  socket_handle,
+                                                  NULL };
 
 static const MVMIOOps op_table = {
     &closable,
