@@ -1,7 +1,5 @@
 #include "moar.h"
 
-#include <math.h>
-
 /* This file contains the "confprog" validator and interpreter.
  *
  * It is there to allow users to give specific behaviour to a few internal
@@ -11,9 +9,9 @@
  * the profiler should start profiling (if it isn't already profiling), and
  * for configuring when the heap snapshot profiler should take a snapshot
  *
- * There is a Perl 6 module that offers a confprog compiler named
+ * There is a Rakudo module that offers a confprog compiler named
  * App::MoarVM::ConfprogCompiler, and nqp's src/vm/moar/HLL/Backend.nqp
- * has a corresponding loader. The nqp and perl6 commandline programs
+ * has a corresponding loader. The nqp and raku commandline programs
  * recognize a --confprog flag.
  */
 
@@ -27,6 +25,26 @@
 #define junkprint(fh, str, ...) do { } while (0)
 #endif
 
+#define CONFPROG_DEBUG_LEVEL_BARE  1
+#define CONFPROG_DEBUG_LEVEL_TRACE 2
+
+// see src/profiler/log.c for the define for debug level 4, which is PROFILER_RESULTS
+
+static void debugprint(MVMuint8 active, MVMThreadContext *tc, const char *str, ...) {
+    va_list args;
+    va_start(args, str);
+
+    if (active) {
+        fprintf(stderr, "%p: ", tc);
+        vfprintf(stderr, str, args);
+        fprintf(stderr, "\n");
+    }
+
+    va_end(args);
+}
+
+#define DEBUG_LVL(level) ((debug_level) & CONFPROG_DEBUG_LEVEL_ ## level)
+
 enum {
     StructSel_Nothing,
     StructSel_Root,
@@ -125,6 +143,9 @@ static MVMuint8 operand_size(MVMThreadContext *tc, MVMuint8 operand) {
 
             case MVM_operand_obj:
             case MVM_operand_type_var: size = 2; break;
+            default:
+                MVM_exception_throw_adhoc(tc, "TODO: unknown operand type");
+                /*fail(val, MSG(val, "unknown operand type %"PRIu32), type);*/
         }
     }
     else {
@@ -190,7 +211,7 @@ static void validate_operands(MVMThreadContext *tc, validatorstate *state) {
     int i = 0;
 
     for (i = 0; i < state->cur_op->num_operands; i++) {
-        validate_operand(tc, state, i, state->cur_op->operands[i]);
+        validate_operand(tc, state, i, operands[i]);
     }
 }
 static void validate_op(MVMThreadContext *tc, validatorstate *state) {
@@ -198,7 +219,7 @@ static void validate_op(MVMThreadContext *tc, validatorstate *state) {
     const MVMOpInfo *info;
     MVMuint8 *prev_op_bc_ptr = state->bc_pointer;
 
-    junkprint(stderr, "validate op at %x\n", state->bc_pointer - state->bytecode_root);
+    junkprint(stderr, "validate op at %lx\n", state->bc_pointer - state->bytecode_root);
 
     if (!MVM_op_is_allowed_in_confprog(opcode)) {
         /*MVMuint16 op;*/
@@ -207,13 +228,20 @@ static void validate_op(MVMThreadContext *tc, validatorstate *state) {
                 /*junkprint(stderr, "op %s (%d) is allowed\n", MVM_op_get_op(op)->name, op);*/
             /*}*/
         /*}*/
+        /* Sigh, so, pointer differences are ptrdiff_t, and (at least on some 32
+         * bit platforms) that's `int`,
+         * (even though on those platforms sizeof(int) == sizeof(long))
+         * and so compilers warn about format string mismatch.
+         * And sigh again, it's not until C11 that we get %tx to specify the
+         * size as that of ptrdiff_t, so we can't rely on on that here.
+         * So cast to unsigned long. */
         MVM_exception_throw_adhoc(tc, "Invalid opcode detected in confprog: %d (%s) at position 0x%lx",
-                opcode, MVM_op_get_op(opcode)->name, state->bc_pointer - state->bytecode_root);
+                opcode, MVM_op_get_op(opcode)->name, (unsigned long) (state->bc_pointer - state->bytecode_root));
     }
     info = MVM_op_get_op(opcode);
     if (!info)
         MVM_exception_throw_adhoc(tc, "Invalid opcode detected in confprog: %d  at position 0x%lx",
-                opcode, state->bc_pointer - state->bytecode_root);
+                opcode, (unsigned long) (state->bc_pointer - state->bytecode_root));
 
     state->prev_op = state->cur_op;
     state->cur_op = info;
@@ -270,7 +298,7 @@ static void validate_op(MVMThreadContext *tc, validatorstate *state) {
                     break;
 
                 default:
-                    MVM_exception_throw_adhoc(tc, "STRUCT_SELECT string length %ld (index %d) NYI or something", string_length, string_idx);
+                    MVM_exception_throw_adhoc(tc, "STRUCT_SELECT string length %"PRIu64" (index %d) NYI or something", string_length, string_idx);
             }
 
             /* Now do a rewrite of const_s into const_i64_16 and noop */
@@ -286,13 +314,10 @@ static void validate_op(MVMThreadContext *tc, validatorstate *state) {
     }
     else if (opcode == MVM_OP_getattr_o) {
         MVMuint8 selected_struct_source = state->selected_struct_source; 
-        MVMuint16 target_reg_num;
         MVMuint16 source_reg_num;
         MVMuint32 string_idx;
-        MVMuint16 hint;
         MVMuint16 *hintptr;
 
-        target_reg_num = *((MVMuint16 *)state->bc_pointer);
         validate_operand(tc, state, 0, state->cur_op->operands[0]);
 
         source_reg_num = *((MVMuint16 *)state->bc_pointer);
@@ -304,7 +329,6 @@ static void validate_op(MVMThreadContext *tc, validatorstate *state) {
         validate_operand(tc, state, 3, state->cur_op->operands[3]);
 
         hintptr = (MVMuint16 *)state->bc_pointer;
-        hint = *hintptr;
         validate_operand(tc, state, 4, state->cur_op->operands[4]);
 
         junkprint(stderr, "currently on %d struct source, string index is %d\n", selected_struct_source, string_idx);
@@ -342,7 +366,7 @@ static void validate_op(MVMThreadContext *tc, validatorstate *state) {
                     }
                 }
                 else {
-                    MVM_exception_throw_adhoc(tc, "STRUCT_SELECT is MVMStaticFrame, no field with length %ld (string heap index %d) implemented", string_length, string_idx);
+                    MVM_exception_throw_adhoc(tc, "STRUCT_SELECT is MVMStaticFrame, no field with length %"PRIu64" (string heap index %d) implemented", string_length, string_idx);
                 }
             }
             else if (selected_struct_source == StructSel_MVMCompUnit) {
@@ -354,7 +378,7 @@ static void validate_op(MVMThreadContext *tc, validatorstate *state) {
                         *hintptr = offsetof(MVMCompUnit, body.hll_name);
                     }
                     else {
-                        MVM_exception_throw_adhoc(tc, "STRUCT_SELECT is MVMCompUnit, no field with length %ld (string heap index %d) implemented", string_length, string_idx);
+                        MVM_exception_throw_adhoc(tc, "STRUCT_SELECT is MVMCompUnit, no field with length %"PRIu64" (string heap index %d) implemented", string_length, string_idx);
                     }
                 }
             }
@@ -380,23 +404,23 @@ static void validate_op(MVMThreadContext *tc, validatorstate *state) {
         new_info = MVM_op_get_op(new_opcode);
         if (!new_info)
             MVM_exception_throw_adhoc(tc, "Invalid opcode detected in confprog: %d  at position 0x%lx",
-                    opcode, state->bc_pointer - state->bytecode_root);
+                    opcode, (unsigned long) (state->bc_pointer - state->bytecode_root));
 
         state->prev_op = state->cur_op;
         state->cur_op = new_info;
 
-        if (new_opcode == MVM_OP_smrt_strify) {
+        if (new_opcode == MVM_OP_unbox_s) {
 
             validate_operand(tc, state, 0, state->cur_op->operands[0]);
             validate_operand(tc, state, 1, state->cur_op->operands[1]);
         }
-        else if (new_opcode == MVM_OP_smrt_intify) {
+        else if (new_opcode == MVM_OP_unbox_i) {
 
             validate_operand(tc, state, 0, state->cur_op->operands[0]);
             validate_operand(tc, state, 1, state->cur_op->operands[1]);
         }
         else {
-            MVM_exception_throw_adhoc(tc, "Confprog: invalid opcode %s followed getcodelocation; only smrt_strify and smrt_intify are allowed.", MVM_op_get_op(new_opcode)->name);
+            MVM_exception_throw_adhoc(tc, "Confprog: invalid opcode %s followed getcodelocation; only unbox_s and unbox_i are allowed.", MVM_op_get_op(new_opcode)->name);
         }
     }
     else {
@@ -405,7 +429,7 @@ static void validate_op(MVMThreadContext *tc, validatorstate *state) {
     state->prev_op_bc = prev_op_bc_ptr;
 }
 
-MVMuint8 MVM_confprog_validate(MVMThreadContext *tc, MVMConfigurationProgram *prog) {
+void MVM_confprog_validate(MVMThreadContext *tc, MVMConfigurationProgram *prog) {
     validatorstate state;
 
     state.confprog = prog;
@@ -452,10 +476,15 @@ static void error_concreteness(MVMThreadContext *tc, MVMObject *object, MVMuint1
             MVM_repr_get_by_id(tc, reprid)->name, purpose, MVM_6model_get_debug_name(tc, object), REPR(object)->name);
 }
 
-MVMint16 stats_position_for_value(MVMThreadContext *tc, MVMuint8 entrypoint, MVMuint64 return_value) {
+static MVMint16 stats_position_for_value(MVMThreadContext *tc, MVMuint8 entrypoint, MVMuint64 return_value) {
     switch (entrypoint) {
         case MVM_PROGRAM_ENTRYPOINT_PROFILER_STATIC:
-            if (return_value >= MVM_CONFPROG_SF_RESULT_TO_BE_DETERMINED && return_value <= MVM_CONFPROG_SF_RESULT_ALWAYS) {
+            if (return_value == MVM_CONFPROG_SF_RESULT_TO_BE_DETERMINED
+                || return_value == MVM_CONFPROG_SF_RESULT_NEVER
+                || return_value == MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_NO
+                || return_value == MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES
+                || return_value == MVM_CONFPROG_SF_RESULT_ALWAYS
+            ) {
                 return return_value;
             }
             if (tc)
@@ -464,12 +493,12 @@ MVMint16 stats_position_for_value(MVMThreadContext *tc, MVMuint8 entrypoint, MVM
         case MVM_PROGRAM_ENTRYPOINT_PROFILER_DYNAMIC:
             if (return_value == 0 || return_value == 1)
                 return MVM_CONFPROG_SF_RESULT_ALWAYS + 1 + return_value;
-            MVM_exception_throw_adhoc(tc, "Can't get stats for out-of-bounds value %ld for dynamic profiler entrypoint", return_value);
+            MVM_exception_throw_adhoc(tc, "Can't get stats for out-of-bounds value %"PRIu64" for dynamic profiler entrypoint", return_value);
             return -1;
         case MVM_PROGRAM_ENTRYPOINT_HEAPSNAPSHOT:
-            if (return_value >= 0 && return_value <= 2)
+            if (return_value <= 2)
                 return MVM_CONFPROG_SF_RESULT_ALWAYS + 1 + 1 + 1 + return_value;
-            MVM_exception_throw_adhoc(tc, "Can't get stats for out-of-bounds value %ld for heapsnapshot entrypoint", return_value);
+            MVM_exception_throw_adhoc(tc, "Can't get stats for out-of-bounds value %"PRIu64" for heapsnapshot entrypoint", return_value);
             return -1;
         default:
             if (tc)
@@ -484,6 +513,8 @@ void MVM_confprog_install(MVMThreadContext *tc, MVMObject *bytecode, MVMObject *
     MVMConfigurationProgram *confprog;
     MVMint16 entrypoint_array[MVM_PROGRAM_ENTRYPOINT_COUNT];
 
+    MVMuint8 debug_level = getenv("MVM_CONFPROG_DEBUG") != NULL ? atoi(getenv("MVM_CONFPROG_DEBUG")) : 0;
+
     CHECK_CONC(bytecode, VMArray, "the bytecode");
     CHECK_CONC(string_array, VMArray, "the string heap");
     CHECK_CONC(entrypoints, VMArray, "the entrypoints list");
@@ -502,12 +533,12 @@ void MVM_confprog_install(MVMThreadContext *tc, MVMObject *bytecode, MVMObject *
         junkprint(stderr, "got a bytecode array with %d (%x) entries\n", bytecode_size, bytecode_size);
 
         if (bytecode_size % 2 == 1) {
-            MVM_exception_throw_adhoc(tc, "installconfprog expected bytecode array to be a multiple of 2 bytes big (got a %ld)",
+            MVM_exception_throw_adhoc(tc, "installconfprog expected bytecode array to be a multiple of 2 bytes big (got a %"PRIu64")",
                     bytecode_size);
         }
 
         if (bytecode_size > 4096) {
-            MVM_exception_throw_adhoc(tc, "confprog too big. maximum 4096, this one has %ld", bytecode_size);
+            MVM_exception_throw_adhoc(tc, "confprog too big. maximum 4096, this one has %"PRIu64"", bytecode_size);
         }
 
         array_contents = ((MVMArray *)bytecode)->body.slots.u8;
@@ -549,7 +580,9 @@ void MVM_confprog_install(MVMThreadContext *tc, MVMObject *bytecode, MVMObject *
         }
     }
 
-    confprog = MVM_calloc(sizeof(MVMConfigurationProgram), 1);
+    confprog = MVM_calloc(1, sizeof(MVMConfigurationProgram));
+
+    confprog->debugging_level = debug_level;
 
     junkprint(stderr, "copying %d (%x) bytecode entries\n", bytecode_size, bytecode_size);
     confprog->bytecode = MVM_malloc(bytecode_size);
@@ -591,6 +624,8 @@ MVMint64 MVM_confprog_run(MVMThreadContext *tc, void *subject, MVMuint8 entrypoi
 
     MVMuint8 *bytecode_start;
 
+    MVMuint8 debug_level = prog->debugging_level;
+
     CPRegister *reg_base = MVM_calloc(prog->reg_count + 1, sizeof(CPRegister));
 
     reg_base[REGISTER_FEATURE_TOGGLE].i64 = initial_feature_value;
@@ -599,11 +634,12 @@ MVMint64 MVM_confprog_run(MVMThreadContext *tc, void *subject, MVMuint8 entrypoi
     cur_op = bytecode_start + prog->entrypoints[entrypoint];
     last_op = bytecode_start + prog->bytecode_length;
 
-    junkprint(stderr, "running confprog for entrypoint %d (at position %d)\n", entrypoint, prog->entrypoints[entrypoint]);
-    junkprint(stderr, "confprog is 0x%x (%d) bytes big", last_op - bytecode_start, last_op - bytecode_start);
-
     MVM_incr(&prog->invoke_counts[entrypoint]);
 
+    debugprint(DEBUG_LVL(BARE), tc, "  running confprog for entrypoint %d (at position 0x%x) (this entrypoint run for the %dth time)\n", entrypoint, prog->entrypoints[entrypoint], MVM_load(&prog->invoke_counts[entrypoint]));
+    /*junkprint(stderr, "confprog is 0x%x (%d) bytes big", last_op - bytecode_start, last_op - bytecode_start);*/
+
+
     runloop: {
         MVMuint16 ins;
 
@@ -618,55 +654,55 @@ MVMint64 MVM_confprog_run(MVMThreadContext *tc, void *subject, MVMuint8 entrypoi
 
         switch (ins) {
             OP(no_op):
-                junkprint(stderr, "noop\n");
+                debugprint(DEBUG_LVL(TRACE), tc, "noop");
                 goto NEXT;
             OP(const_i64):
                 GET_REG(cur_op, 0).i64 = MVM_BC_get_I64(cur_op, 2);
-                junkprint(stderr, "const_i64 %d into %d\n", MVM_BC_get_I64(cur_op, 2), GET_UI16(cur_op, 0));
+                debugprint(DEBUG_LVL(TRACE), tc, "const_i64 %ld into %d", MVM_BC_get_I64(cur_op, 2), GET_UI16(cur_op, 0));
                 cur_op += 10;
                 goto NEXT;
             OP(const_n64):
                 GET_REG(cur_op, 0).n64 = MVM_BC_get_N64(cur_op, 2);
-                junkprint(stderr, "const_n64 %f into %d\n", MVM_BC_get_N64(cur_op, 2), GET_UI16(cur_op, 0));
+                debugprint(DEBUG_LVL(TRACE), tc, "const_n64 %f into %d", MVM_BC_get_N64(cur_op, 2), GET_UI16(cur_op, 0));
                 cur_op += 10;
                 goto NEXT;
             OP(const_i64_16):
                 GET_REG(cur_op, 0).i64 = GET_I16(cur_op, 2);
-                junkprint(stderr, "const_i64_16 %d into %d\n", GET_I16(cur_op, 2), GET_UI16(cur_op, 0));
+                debugprint(DEBUG_LVL(TRACE), tc, "const_i64_16 %d into %d", GET_I16(cur_op, 2), GET_UI16(cur_op, 0));
                 cur_op += 4;
                 goto NEXT;
             OP(const_s):
                 GET_REG(cur_op, 0).s = MVM_repr_at_pos_s(tc, prog->string_heap, GET_UI32(cur_op, 2));
-                junkprint(stderr, "const_s into reg %d\n", GET_UI16(cur_op, 0));
+                debugprint(DEBUG_LVL(TRACE), tc, "const_s into reg %d", GET_UI16(cur_op, 0));
                 cur_op += 6;
                 goto NEXT;
             OP(set):
                 GET_REG(cur_op, 0) = GET_REG(cur_op, 2);
-                junkprint(stderr, "set\n");
+                debugprint(DEBUG_LVL(TRACE), tc, "set %d <- %d", GET_UI16(cur_op, 0), GET_UI16(cur_op, 2));
                 cur_op += 4;
                 goto NEXT;
             OP(say): {
                 MVMString *s = GET_REG(cur_op, 0).s;
+                debugprint(DEBUG_LVL(TRACE), tc, "say %d", GET_UI16(cur_op, 0));
                 if (s && IS_CONCRETE(s))
                     MVM_string_say(tc, s);
-                junkprint(stderr, "say\n");
                 cur_op += 2;
                 goto NEXT;
             }
             OP(getattr_o):
-                junkprint(stderr, "struct select: %x\n", reg_base[REGISTER_STRUCT_SELECT].i64);
+                debugprint(DEBUG_LVL(TRACE), tc, "struct select: %lx", (unsigned long)reg_base[REGISTER_STRUCT_SELECT].i64);
                 if (reg_base[REGISTER_STRUCT_SELECT].i64 == StructSel_Nothing) {
-                    junkprint(stderr, "getattr_o into %d\n", GET_UI16(cur_op, 0));
+                    debugprint(DEBUG_LVL(TRACE), tc, "getattr_o into %d", GET_UI16(cur_op, 0));
                 }
                 else if (reg_base[REGISTER_STRUCT_SELECT].i64 == StructSel_Root) {
                     MVMuint16 field_select = GET_UI16(cur_op, 10);
                     if (field_select == FieldSel_staticframe) {
                         reg_base[REGISTER_STRUCT_ACCUMULATOR].any = tc->cur_frame->static_info;
-                        junkprint(stderr, "get a static frame into the struct accumulator: %x\n", tc->cur_frame->static_info);
+                        debugprint(DEBUG_LVL(TRACE), tc, "get a static frame into the struct accumulator: %p", (void *)tc->cur_frame->static_info);
                     }
                     else if (field_select == FieldSel_frame) {
                         reg_base[REGISTER_STRUCT_ACCUMULATOR].any = tc->cur_frame;
-                        junkprint(stderr, "get a frame into the struct accumulator: %x\n", tc->cur_frame);
+                        debugprint(DEBUG_LVL(TRACE), tc, "get a frame into the struct accumulator: %p", (void *)tc->cur_frame);
                     }
                     else {
                         fprintf(stderr, "NYI case of getattr_o on root struct hit\n");
@@ -674,19 +710,19 @@ MVMint64 MVM_confprog_run(MVMThreadContext *tc, void *subject, MVMuint8 entrypoi
                     }
                 }
                 else {
-                    junkprint(stderr, "getting the struct accumulator's field at offset 0x%x into register %d\n",
+                    debugprint(DEBUG_LVL(TRACE), tc, "getting the struct accumulator's field at offset 0x%x into register %d",
                             GET_UI16(cur_op, 10), GET_UI16(cur_op, 0));
-                    junkprint(stderr, "register %d contents before: 0x%x\n", GET_UI16(cur_op, 0), reg_base[GET_UI16(cur_op, 0)].any);
-                    junkprint(stderr, "register %d contents before: 0x%x\n", REGISTER_STRUCT_ACCUMULATOR, reg_base[REGISTER_STRUCT_ACCUMULATOR].any);
+                    debugprint(DEBUG_LVL(TRACE), tc, "register %d contents before: 0x%p", GET_UI16(cur_op, 0), reg_base[GET_UI16(cur_op, 0)].any);
+                    debugprint(DEBUG_LVL(TRACE), tc, "register %d contents before: 0x%p", REGISTER_STRUCT_ACCUMULATOR, reg_base[REGISTER_STRUCT_ACCUMULATOR].any);
                     reg_base[GET_UI16(cur_op, 0)].any =
                         *((void **)(((char *)reg_base[REGISTER_STRUCT_ACCUMULATOR].any) + GET_UI16(cur_op, 10)));
-                    junkprint(stderr, "register %d contents now: 0x%x\n", GET_UI16(cur_op, 0), reg_base[GET_UI16(cur_op, 0)].any);
+                    debugprint(DEBUG_LVL(TRACE), tc, "register %d contents now: 0x%p", GET_UI16(cur_op, 0), reg_base[GET_UI16(cur_op, 0)].any);
                     /*(char *)(&reg_base[REGISTER_STRUCT_ACCUMULATOR]) = (char *)*/
                 }
                 cur_op += 12;
                 goto NEXT;
             OP(eq_s):
-                junkprint(stderr, "eq_s with %d and %d\n", GET_UI16(cur_op, 2), GET_UI16(cur_op, 4));
+                debugprint(DEBUG_LVL(TRACE), tc, "eq_s with %d and %d", GET_UI16(cur_op, 2), GET_UI16(cur_op, 4));
                 GET_REG(cur_op, 0).i64 = MVM_string_equal(tc,
                     GET_REG(cur_op, 2).s, GET_REG(cur_op, 4).s);
                 cur_op += 6;
@@ -788,22 +824,19 @@ MVMint64 MVM_confprog_run(MVMThreadContext *tc, void *subject, MVMuint8 entrypoi
                 GET_REG(cur_op, 0).n64 = MVM_coerce_s_n(tc, GET_REG(cur_op, 2).s);
                 cur_op += 4;
                 goto NEXT;
-            OP(smrt_strify): {
-                /* Increment PC before calling coercer, as it may make
-                 * a method call to get the result. */
-                MVMObject   *obj = GET_REG(cur_op, 2).o;
-                MVMRegister *res = (MVMRegister *)&GET_REG(cur_op, 0);
-                cur_op += 4;
-                MVM_coerce_smart_stringify(tc, obj, res);
-                goto NEXT;
-            }
             OP(gt_n):
+                debugprint(DEBUG_LVL(TRACE), tc, "%f > %f into %d", GET_REG(cur_op, 2).n64, GET_REG(cur_op, 4).n64, GET_UI16(cur_op, 0));
                 GET_REG(cur_op, 0).i64 = GET_REG(cur_op, 2).n64 >  GET_REG(cur_op, 4).n64;
                 cur_op += 6;
                 goto NEXT;
+            OP(ge_n):
+                debugprint(DEBUG_LVL(TRACE), tc, "%f >= %f into %d", GET_REG(cur_op, 2).n64, GET_REG(cur_op, 4).n64, GET_UI16(cur_op, 0));
+                GET_REG(cur_op, 0).i64 = GET_REG(cur_op, 2).n64 >=  GET_REG(cur_op, 4).n64;
+                cur_op += 6;
+                goto NEXT;
             OP(rand_n):
                 GET_REG(cur_op, 0).n64 = MVM_proc_rand_n(tc);
-                junkprint(stderr, "rand_n result %f into %d\n", GET_REG(cur_op, 0).n64, GET_UI16(cur_op, 0));
+                debugprint(DEBUG_LVL(TRACE), tc, "rand_n result %f into %d", GET_REG(cur_op, 0).n64, GET_UI16(cur_op, 0));
                 cur_op += 2;
                 goto NEXT;
             OP(goto):
@@ -815,21 +848,20 @@ MVMint64 MVM_confprog_run(MVMThreadContext *tc, void *subject, MVMuint8 entrypoi
                 else
                     cur_op += 6;
                 goto NEXT;
-            OP(time_i):
-                GET_REG(cur_op, 0).i64 = MVM_proc_time_i(tc);
+            OP(time):
+                GET_REG(cur_op, 0).u64 = MVM_proc_time(tc);
                 cur_op += 2;
                 goto NEXT;
             OP(exit): {
-                MVMint64 exit_code = GET_REG(cur_op, 0).i64;
                 goto finish_main_loop;
             OP(index_s):
-                junkprint(stderr, "index_s into %d with %d and %d starting at %d\n",
+                debugprint(DEBUG_LVL(TRACE), tc, "index_s into %d with %d and %d starting at %d",
                         GET_UI16(cur_op, 0), GET_UI16(cur_op, 2), GET_UI16(cur_op, 4), GET_UI16(cur_op, 6));
-                junkprint(stderr, "values %p and %p starting at %d\n",
+                debugprint(DEBUG_LVL(TRACE), tc, "values %p and %p starting at %ld",
                         GET_REG(cur_op, 2).s, GET_REG(cur_op, 4).s, GET_REG(cur_op, 6).i64);
                 GET_REG(cur_op, 0).i64 = MVM_string_index(tc,
                     STRING_OR_EMPTY(GET_REG(cur_op, 2).s), STRING_OR_EMPTY(GET_REG(cur_op, 4).s), GET_REG(cur_op, 6).i64);
-                junkprint(stderr, "index_s result: %d\n", GET_REG(cur_op, 0).i64);
+                debugprint(DEBUG_LVL(TRACE), tc, "index_s result: %ld", GET_REG(cur_op, 0).i64);
                 cur_op += 8;
                 goto NEXT;
             OP(eqat_s):
@@ -920,7 +952,7 @@ MVMint64 MVM_confprog_run(MVMThreadContext *tc, void *subject, MVMuint8 entrypoi
                 cur_op += 6;
                 goto NEXT;
             OP(getcodelocation): {
-                MVMuint32 line_out = 0;
+                MVMint32 line_out = 0;
                 MVMString *file_out = NULL;
                 MVMObject *code_obj = (MVMObject *)((MVMStaticFrame *)reg_base[REGISTER_STRUCT_ACCUMULATOR].any)->body.static_code;
                 cur_op += 4;
@@ -928,7 +960,8 @@ MVMint64 MVM_confprog_run(MVMThreadContext *tc, void *subject, MVMuint8 entrypoi
                 ins = *((MVMuint16 *)cur_op);
                 cur_op += 2;
                 MVM_code_location_out(tc, code_obj, &file_out, &line_out);
-                if (ins == MVM_OP_smrt_strify) {
+                MVM_exception_throw_adhoc(tc, "getcodelocation in conf prog needs updates after new-disp");
+                if (ins == MVM_OP_unbox_s) {
                     GET_REG(cur_op, 0).s = file_out ? file_out : tc->instance->str_consts.empty;
                 }
                 else {
@@ -955,11 +988,52 @@ finish_main_loop:
     stats_slot = stats_position_for_value(NULL, entrypoint, result);
 
     if (stats_slot != -1) {
-        MVM_store(&prog->last_return_time[stats_slot], (AO_t)MVM_proc_time_n(tc));
+        MVM_store(&prog->last_return_time[stats_slot], AO_CAST(MVM_proc_time(tc)));
         MVM_incr(&prog->return_counts[stats_slot]);
     }
 
-    junkprint(stderr, "confprog result value: %lld\n\n", result);
+    {
+        const char *resultname = "unknown result meaning";
+        switch (entrypoint) {
+            case MVM_PROGRAM_ENTRYPOINT_PROFILER_STATIC:
+            case MVM_PROGRAM_ENTRYPOINT_PROFILER_DYNAMIC:
+                switch (result) {
+                    case MVM_CONFPROG_SF_RESULT_TO_BE_DETERMINED:
+                        resultname = "RESULT_TO_BE_DETERMINED";
+                        break;
+                    case MVM_CONFPROG_SF_RESULT_NEVER:
+                        resultname = "RESULT_NEVER";
+                        break;
+                    case MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_NO:
+                        resultname = "RESULT_DYNAMIC_SUGGEST_NO";
+                        break;
+                    case MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES:
+                        resultname = "RESULT_DYNAMIC_SUGGEST_YES";
+                        break;
+                    case MVM_CONFPROG_SF_RESULT_ALWAYS:
+                        resultname = "RESULT_ALWAYS";
+                        break;
+                }
+                break;
+            case MVM_PROGRAM_ENTRYPOINT_HEAPSNAPSHOT:
+                switch (result) {
+                    case MVM_CONFPROG_HEAPSNAPSHOT_RESULT_NOTHING:
+                        resultname = "RESULT_NOTHING";
+                        break;
+                    case MVM_CONFPROG_HEAPSNAPSHOT_RESULT_SNAPSHOT:
+                        resultname = "RESULT_SNAPSHOT";
+                        break;
+                    case MVM_CONFPROG_HEAPSNAPSHOT_RESULT_STATS:
+                        resultname = "RESULT_STATS";
+                        break;
+                    case MVM_CONFPROG_HEAPSNAPSHOT_RESULT_SNAPSHOT_WITH_STATS:
+                        resultname = "RESULT_SNAPSHOT_WITH_STATS";
+                        break;
+                }
+                break;
+        }
+        debugprint(DEBUG_LVL(BARE), tc, "confprog result value: %"PRIi64" (%s)\n", result, resultname);
+    }
 
     return result;
 }
