@@ -1,5 +1,22 @@
 #include "moar.h"
 
+#define CONFPROG_DEBUG_LEVEL_PROFILER_RESULTS 4
+
+static void debugprint(MVMuint8 active, MVMThreadContext *tc, const char *str, ...) {
+    va_list args;
+    va_start(args, str);
+
+    if (active) {
+        fprintf(stderr, "%p: ", tc);
+        vfprintf(stderr, str, args);
+        fprintf(stderr, "\n");
+    }
+
+    va_end(args);
+}
+
+#define DEBUG_LVL(level) ((debug_level) & CONFPROG_DEBUG_LEVEL_ ## level)
+
 /* Gets the current thread's profiling data structure, creating it if needed. */
 static MVMProfileThreadData * get_thread_data(MVMThreadContext *tc) {
     if (!tc->prof_data) {
@@ -46,42 +63,58 @@ void MVM_profile_log_enter(MVMThreadContext *tc, MVMStaticFrame *sf, MVMuint64 m
         MVMuint32 i;
         if (ptd->current_call)
             for (i = 0; i < ptd->current_call->num_succ; i++)
-                if (ptd->current_call->succ[i]->sf == sf)
+                if (ptd->staticframe_array[ptd->current_call->succ[i]->sf_idx] == sf)
                     pcn = ptd->current_call->succ[i];
 
         if (MVM_UNLIKELY(!ptd->current_call)) {
+            /* debug_level used by the debugprint macros */
+            MVMuint8 debug_level = tc->instance->confprog ? tc->instance->confprog->debugging_level : 0;
             MVMuint8 has_confprogs = tc->instance->confprog && (MVM_confprog_has_entrypoint(tc, MVM_PROGRAM_ENTRYPOINT_PROFILER_STATIC)
                         || MVM_confprog_has_entrypoint(tc, MVM_PROGRAM_ENTRYPOINT_PROFILER_STATIC));
-            /*fprintf(stderr, "\n\n no current call exists \n");*/
             /* At the very beginning of profiling with a confprog set, the
              * root call_graph node needs to be created. */
             if (has_confprogs && !ptd->call_graph) {
-                /*fprintf(stderr, "initialized initial call graph node of thread %p\n", tc);*/
-                ptd->call_graph = MVM_calloc(sizeof(MVMProfileCallNode), 1);
+                debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "initialized initial call graph node\n");
+                ptd->call_graph = MVM_calloc(1, sizeof(MVMProfileCallNode));
             }
             /* In that case, we've got to check if the SF in question is
              * desired as an entry point */
-            if (sf->body.instrumentation && tc->instance->confprog) {
+            if (sf->body.instrumentation && has_confprogs) {
                 MVMStaticFrameInstrumentation *instrumentation = sf->body.instrumentation;
                 if (instrumentation->profiler_confprog_result == MVM_CONFPROG_SF_RESULT_NEVER) {
-                    /*fprintf(stderr, "encountered a 'never profile' SF\n");*/
                     goto confprog_refused_enter;
                 }
                 else if (instrumentation->profiler_confprog_result == MVM_CONFPROG_SF_RESULT_TO_BE_DETERMINED) {
                     if (MVM_confprog_has_entrypoint(tc, MVM_PROGRAM_ENTRYPOINT_PROFILER_STATIC)) {
                         MVMuint8 result;
-                        /*fprintf(stderr, "confprog setting was already 'to be determined'\n");*/
+                        debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "running 'profiler_static' entrypoint in confprog");
                         result = MVM_confprog_run(tc, (void *)sf, MVM_PROGRAM_ENTRYPOINT_PROFILER_STATIC, MVM_CONFPROG_SF_RESULT_ALWAYS);
                         instrumentation->profiler_confprog_result = result;
                         if (result == MVM_CONFPROG_SF_RESULT_NEVER) {
-                            /*fprintf(stderr, "configured a 'never profile' SF with confprog\n");*/
+                            debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog for SF resulted in 'never profile'");
                             goto confprog_refused_enter;
                         }
-                        /*fprintf(stderr, "configured a %d SF\n", result);*/
+                        if (DEBUG_LVL(PROFILER_RESULTS)) {
+                            switch (result) {
+                                case MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES:
+                                case MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_NO:
+                                    debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog result: run dynamic program with default value '%s' (result value: %d)", result == MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES ? "yes" : "no", result);
+                                    break;
+                                case MVM_CONFPROG_SF_RESULT_ALWAYS:
+                                    debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog result: always profile from this SF (result value: %d)", result);
+                                    break;
+                                case MVM_CONFPROG_SF_RESULT_TO_BE_DETERMINED:
+                                    debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog result: to be determined (result value %d) - will enter this time, but re-run next time", result);
+                                    break;
+                                default:
+                                    debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  unrecognized result value from confprog: %d", result);
+                                    break;
+                            }
+                            debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog for SF resulted in 'never profile'");
+                        }
                     }
                     else if (MVM_confprog_has_entrypoint(tc, MVM_PROGRAM_ENTRYPOINT_PROFILER_DYNAMIC)) {
                         instrumentation->profiler_confprog_result = MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES;
-                        /*fprintf(stderr, "configured a 'consider dynamic confprog' SF (because no static prog installed)\n");*/
                     }
                     else {
                         MVM_oops(tc, "here we are, what now?");
@@ -89,22 +122,21 @@ void MVM_profile_log_enter(MVMThreadContext *tc, MVMStaticFrame *sf, MVMuint64 m
                 }
                 if (instrumentation->profiler_confprog_result == MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_NO
                         || instrumentation->profiler_confprog_result == MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES) {
-                    /*fprintf(stderr, "confprog result is 'please run dynamic code'\n");*/
+                    debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "running 'profiler_dynamic' entrypoint in confprog\n");
                     if (MVM_confprog_has_entrypoint(tc, MVM_PROGRAM_ENTRYPOINT_PROFILER_DYNAMIC)) {
                         if (!MVM_confprog_run(tc, (void *)tc->cur_frame, MVM_PROGRAM_ENTRYPOINT_PROFILER_DYNAMIC, instrumentation->profiler_confprog_result == MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES)) {
-                            /*fprintf(stderr, "ran a dynamic confprog for frame, but it said no\n");*/
+                            debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog result: no.\n");
                             goto confprog_refused_enter;
                         }
-                        /*fprintf(stderr, "ran a dynamic confprog for frame, and it said yes\n");*/
+                        debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog result: yes.\n");
                     }
                     else {
-                        /*fprintf(stderr, "sf result was to consider a dynamic confprog, but none was installed\n");*/
+                        debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  static confprog said to run dynamic confprog, but none is installed - not profiling.\n");
                         goto confprog_refused_enter;
                     }
                 }
                 was_entered_via_confprog = 1;
             }
-            /*fprintf(stderr, "%p entered %p via confprog; nctd is %d\n", tc, sf, ptd->non_calltree_depth);*/
         }
         /*else {*/
             /*fprintf(stderr, "there actually was a current_call. also, pcn is %p\n", pcn);*/
@@ -113,10 +145,19 @@ void MVM_profile_log_enter(MVMThreadContext *tc, MVMStaticFrame *sf, MVMuint64 m
         /* If we didn't find a call graph node, then create one and add it to the
          * graph. */
         if (!pcn) {
+            MVMuint32 search;
             if (was_entered_via_confprog)
                 ptd->current_call = ptd->call_graph;
             pcn = make_new_pcn(ptd, current_hrtime);
-            pcn->sf = sf;
+            for (search = 0; search < MVM_VECTOR_ELEMS(ptd->staticframe_array); search++) {
+                if (ptd->staticframe_array[search] == sf) {
+                    break;
+                }
+            }
+            if (search == MVM_VECTOR_ELEMS(ptd->staticframe_array)) {
+                MVM_VECTOR_PUSH(ptd->staticframe_array, sf);
+            }
+            pcn->sf_idx = search;
         }
 
         /* Increment entry counts. */
@@ -165,7 +206,7 @@ void MVM_profile_log_enter_native(MVMThreadContext *tc, MVMObject *nativecallsit
     callbody = MVM_nativecall_get_nc_body(tc, nativecallsite);
     if (ptd->current_call)
         for (i = 0; i < ptd->current_call->num_succ; i++)
-            if (ptd->current_call->succ[i]->sf == NULL)
+            if (ptd->staticframe_array[ptd->current_call->succ[i]->sf_idx] == NULL)
                 if (strcmp(callbody->sym_name,
                            ptd->current_call->succ[i]->native_target_name) == 0) {
                     pcn = ptd->current_call->succ[i];
@@ -235,7 +276,7 @@ void MVM_profile_log_unwind(MVMThreadContext *tc) {
             return;
         lpcn = pcn;
         log_exit(tc, 1);
-    } while (lpcn->sf != tc->cur_frame->static_info);
+    } while (ptd->staticframe_array[lpcn->sf_idx] != tc->cur_frame->static_info);
 }
 
 /* Called when we take a continuation. Leaves the static frames from the point
@@ -262,13 +303,13 @@ MVMProfileContinuationData * MVM_profile_log_continuation_control(MVMThreadConte
                 sfs        = MVM_realloc(sfs, alloc_sfs * sizeof(MVMStaticFrame *));
                 modes      = MVM_realloc(modes, alloc_sfs * sizeof(MVMuint64));
             }
-            sfs[num_sfs]   = pcn->sf;
+            sfs[num_sfs]   = ptd->staticframe_array[pcn->sf_idx];
             modes[num_sfs] = pcn->entry_mode;
             num_sfs++;
 
             lpcn = pcn;
             log_exit(tc, 1);
-        } while (lpcn->sf != cur_frame->static_info);
+        } while (ptd->staticframe_array[lpcn->sf_idx] != cur_frame->static_info);
 
         last_frame = cur_frame;
         cur_frame = cur_frame->caller;
@@ -295,7 +336,7 @@ void MVM_profile_log_thread_created(MVMThreadContext *tc, MVMThreadContext *chil
 }
 
 /* Logs one allocation, potentially scalar replaced. */
-void log_one_allocation(MVMThreadContext *tc, MVMObject *obj, MVMProfileCallNode *pcn, MVMuint8 replaced) {
+static void log_one_allocation(MVMThreadContext *tc, MVMObject *obj, MVMProfileCallNode *pcn, MVMuint8 replaced) {
     MVMObject *what = STABLE(obj)->WHAT;
     MVMuint32 i;
     MVMuint8 allocation_target;
@@ -312,7 +353,7 @@ void log_one_allocation(MVMThreadContext *tc, MVMObject *obj, MVMProfileCallNode
 
     /* See if there's an existing node to update. */
     for (i = 0; i < pcn->num_alloc; i++) {
-        if (pcn->alloc[i].type == what) {
+        if (tc->prof_data->type_array[pcn->alloc[i].type_idx] == what) {
             if (allocation_target == 0)
                 pcn->alloc[i].allocations_interp++;
             else if (allocation_target == 1)
@@ -327,11 +368,30 @@ void log_one_allocation(MVMThreadContext *tc, MVMObject *obj, MVMProfileCallNode
 
     /* No entry; create one. */
     if (pcn->num_alloc == pcn->alloc_alloc) {
-        pcn->alloc_alloc += 8;
-        pcn->alloc = MVM_realloc(pcn->alloc,
-            pcn->alloc_alloc * sizeof(MVMProfileAllocationCount));
+        if (pcn->alloc_alloc == 0) {
+            pcn->alloc_alloc++;
+            pcn->alloc = MVM_malloc(pcn->alloc_alloc * sizeof(MVMProfileAllocationCount));
+        }
+        else {
+            pcn->alloc_alloc *= 2;
+            pcn->alloc = MVM_realloc(
+                    pcn->alloc,
+                    pcn->alloc_alloc * sizeof(MVMProfileAllocationCount));
+        }
     }
-    pcn->alloc[pcn->num_alloc].type        = what;
+    {
+        MVMuint32 search;
+        for (search = 0; search < MVM_VECTOR_ELEMS(tc->prof_data->type_array); search++) {
+            if (tc->prof_data->type_array[search] == what) {
+                break;
+            }
+        }
+        if (search == MVM_VECTOR_ELEMS(tc->prof_data->type_array)) {
+            MVM_VECTOR_PUSH(tc->prof_data->type_array, what);
+        }
+        pcn->alloc[pcn->num_alloc].type_idx    = search;
+    }
+
     pcn->alloc[pcn->num_alloc].allocations_interp = allocation_target == 0;
     pcn->alloc[pcn->num_alloc].allocations_spesh  = allocation_target == 1;
     pcn->alloc[pcn->num_alloc].allocations_jit    = allocation_target == 2;
@@ -371,14 +431,14 @@ void MVM_profiler_log_gc_deallocate(MVMThreadContext *tc, MVMObject *object) {
 
         MVMuint8 dealloc_target = 0;
 
-        if (what->header.flags & MVM_CF_FORWARDER_VALID)
+        if (what->header.flags2 & MVM_CF_FORWARDER_VALID)
             what = (MVMObject *)what->header.sc_forward_u.forwarder;
 
         MVM_ASSERT_NOT_FROMSPACE(tc, what);
 
-        if (item->flags & MVM_CF_SECOND_GEN)
+        if (item->flags2 & MVM_CF_SECOND_GEN)
             dealloc_target = 2;
-        else if (item->flags & MVM_CF_NURSERY_SEEN)
+        else if (item->flags2 & MVM_CF_NURSERY_SEEN)
             dealloc_target = 1;
 
         /* See if there's an existing node to update. */
@@ -396,9 +456,16 @@ void MVM_profiler_log_gc_deallocate(MVMThreadContext *tc, MVMObject *object) {
 
         /* No entry; create one. */
         if (pgc->num_dealloc == pgc->alloc_dealloc) {
-            pgc->alloc_dealloc += 8;
-            pgc->deallocs = MVM_realloc(pgc->deallocs,
-                pgc->alloc_dealloc * sizeof(MVMProfileDeallocationCount));
+            if (pgc->alloc_dealloc == 0) {
+                pgc->alloc_dealloc++;
+                pgc->deallocs = MVM_malloc(pgc->alloc_dealloc * sizeof(MVMProfileDeallocationCount));
+            }
+            else {
+                pgc->alloc_dealloc *= 2;
+                pgc->deallocs = MVM_realloc(
+                        pgc->deallocs,
+                        pgc->alloc_dealloc * sizeof(MVMProfileDeallocationCount));
+            }
         }
         pgc->deallocs[pgc->num_dealloc].type                   = what;
         pgc->deallocs[pgc->num_dealloc].deallocs_nursery_fresh = dealloc_target == 0;
@@ -441,6 +508,7 @@ void MVM_profiler_log_gc_start(MVMThreadContext *tc, MVMuint32 full, MVMuint32 t
     gc->num_dealloc = 0;
     gc->alloc_dealloc = 0;
     gc->deallocs = NULL;
+    gc->num_stolen_gen2roots = 0;
 
     /* Record start time. */
     ptd->cur_gc_start_time = uv_hrtime();
@@ -488,6 +556,14 @@ void MVM_profiler_log_unmanaged_data_promoted(MVMThreadContext *tc, MVMuint64 am
     ptd->gc_promoted_unmanaged_bytes += amount;
 }
 
+void MVM_profiler_log_gen2_roots(MVMThreadContext *tc, MVMuint64 amount, MVMThreadContext *other) {
+    if (tc != other) {
+        MVMProfileThreadData *ptd = get_thread_data(tc);
+
+        ptd->gcs[ptd->num_gcs].num_stolen_gen2roots += amount;
+    }
+}
+
 /* Log that we're starting some work on bytecode specialization or JIT. */
 void MVM_profiler_log_spesh_start(MVMThreadContext *tc) {
     /* Record start time. */
