@@ -49,9 +49,9 @@ static const MVMuint8 utf8d[] = {
   12,36,12,12,12,12,12,12,12,12,12,12,
 };
 
-static MVMint32
+MVM_STATIC_INLINE MVMint32
 decode_utf8_byte(MVMint32 *state, MVMGrapheme32 *codep, MVMuint8 byte) {
-  MVMint32 type = utf8d[byte];
+  const MVMint32 type = utf8d[byte];
 
   *codep = (*state != UTF8_ACCEPT) ?
     (byte & 0x3fu) | (*codep << 6) :
@@ -166,7 +166,69 @@ static MVMint32 utf8_encode(MVMuint8 *bp, MVMCodepoint cp) {
 
  /* end not_gerd section */
 
-#define UTF8_MAXINC (32 * 1024 * 1024)
+static void encoding_error(MVMThreadContext *tc, const MVMuint8 *bytes, int error_pos, int line, int col) {
+                    /* Different error messages for the first few bytes, so that
+                     * we print as much context as possible */
+    char location[sizeof(" at line  col ") + 2 * (3 * sizeof(int) * CHAR_BIT/8 + 2)] = "";
+    if (line >= 0)
+        sprintf(location, " at line %u col %u", line, col);
+
+    if (error_pos >= 3) {
+        MVMuint8 a = bytes[error_pos - 2], b = bytes[error_pos - 1], c = bytes[error_pos];
+        MVM_exception_throw_adhoc(tc, "Malformed UTF-8 near bytes %02hhx %02hhx %02hhx%s", a, b, c, location);
+    }
+    else if (error_pos == 2) {
+        MVMuint8 a = bytes[error_pos - 1], b = bytes[error_pos];
+        MVM_exception_throw_adhoc(tc, "Malformed UTF-8 near bytes %02hhx %02hhx%s", a, b, location);
+    }
+    else if (error_pos == 1) {
+        MVMuint8 a = bytes[error_pos];
+        MVM_exception_throw_adhoc(tc, "Malformed UTF-8 near byte %02hhx%s", a, location);
+    }
+    else {
+        MVM_exception_throw_adhoc(tc, "Malformed UTF-8%s", location);
+    }
+}
+
+/* Decodes the specified number of bytes of utf8 into an NFG string, expecting to get
+ * an error, and then reporting where the error occured. */
+void utf8_decode_errors(MVMThreadContext *tc, const char *utf8, size_t bytes) {
+    MVMint32 line_ending = 0;
+    MVMint32 state = 0;
+    MVMint32 line  = 1;
+    MVMint32 col   = 1;
+    MVMCodepoint codepoint;
+    size_t orig_bytes = bytes;
+    const char *orig_utf8 = utf8;
+
+    for (; bytes; ++utf8, --bytes) {
+        switch(decode_utf8_byte(&state, &codepoint, (MVMuint8)*utf8)) {
+        case UTF8_ACCEPT:
+            /* this could be reorganized into several nested ugly if/else :/ */
+            if (!line_ending && (codepoint == 10 || codepoint == 13)) {
+                /* Detect the style of line endings.
+                 * Select whichever comes first.
+                 * First or only part of first line ending. */
+                line_ending = codepoint;
+                col = 1; line++;
+            }
+            else if (line_ending && codepoint == line_ending) {
+                /* first or only part of next line ending */
+                col = 1; line++;
+            }
+            else if (codepoint == 10 || codepoint == 13) {
+                /* second part of line ending; ignore */
+            }
+            else /* non-line ending codepoint */
+                col++;
+            break;
+        case UTF8_REJECT: {
+            encoding_error(tc, (MVMuint8 *)orig_utf8, orig_bytes - bytes, line, col);
+            break;
+        }
+        }
+    }
+}
 
 /* Decodes the specified number of bytes of utf8 into an NFG string, creating
  * a result of the specified type. The type must have the MVMString REPR. */
@@ -174,37 +236,27 @@ MVMString * MVM_string_utf8_decode(MVMThreadContext *tc, const MVMObject *result
     MVMString *result = (MVMString *)REPR(result_type)->allocate(tc, STABLE(result_type));
     MVMint32 count = 0;
     MVMCodepoint codepoint;
-    MVMint32 line_ending = 0;
     MVMint32 state = 0;
     MVMint32 bufsize = bytes;
     MVMGrapheme32 *buffer = MVM_malloc(sizeof(MVMGrapheme32) * bufsize);
-    size_t orig_bytes;
-    const char *orig_utf8;
-    MVMint32 line;
-    MVMint32 col;
+    size_t orig_bytes = bytes;
+    const char *orig_utf8 = utf8;
     MVMint32 ready;
 
     /* Need to normalize to NFG as we decode. */
     MVMNormalizer norm;
     MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFG);
 
-    orig_bytes = bytes;
-    orig_utf8 = utf8;
-
     for (; bytes; ++utf8, --bytes) {
-        switch(decode_utf8_byte(&state, &codepoint, (MVMuint8)*utf8)) {
+        switch(MVM_EXPECT(decode_utf8_byte(&state, &codepoint, (MVMuint8)*utf8), UTF8_ACCEPT)) {
         case UTF8_ACCEPT: { /* got a codepoint */
             MVMGrapheme32 g;
             ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, codepoint, &g);
             if (ready) {
-                while (count + ready >= bufsize) { /* if the buffer's full make a bigger one */
-                    buffer = MVM_realloc(buffer, sizeof(MVMGrapheme32) * (
-                        bufsize >= UTF8_MAXINC ? (bufsize += UTF8_MAXINC) : (bufsize *= 2)
-                    ));
-                }
                 buffer[count++] = g;
-                while (--ready > 0)
+                while (--ready > 0) {
                     buffer[count++] = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+                }
             }
             break;
         }
@@ -212,35 +264,8 @@ MVMString * MVM_string_utf8_decode(MVMThreadContext *tc, const MVMObject *result
             /* found a malformed sequence; parse it again this time tracking
              * line and col numbers. */
             MVM_unicode_normalizer_cleanup(tc, &norm); /* Since we'll throw. */
-            bytes = orig_bytes; utf8 = orig_utf8; state = 0; line = 1; col = 1;
-            for (; bytes; ++utf8, --bytes) {
-                switch(decode_utf8_byte(&state, &codepoint, (MVMuint8)*utf8)) {
-                case UTF8_ACCEPT:
-                    /* this could be reorganized into several nested ugly if/else :/ */
-                    if (!line_ending && (codepoint == 10 || codepoint == 13)) {
-                        /* Detect the style of line endings.
-                         * Select whichever comes first.
-                         * First or only part of first line ending. */
-                        line_ending = codepoint;
-                        col = 1; line++;
-                    }
-                    else if (line_ending && codepoint == line_ending) {
-                        /* first or only part of next line ending */
-                        col = 1; line++;
-                    }
-                    else if (codepoint == 10 || codepoint == 13) {
-                        /* second part of line ending; ignore */
-                    }
-                    else /* non-line ending codepoint */
-                        col++;
-                    break;
-                case UTF8_REJECT:
-                    MVM_free(buffer);
-                    MVM_exception_throw_adhoc(tc, "Malformed UTF-8 at line %u col %u", line, col);
-                }
-            }
             MVM_free(buffer);
-            MVM_exception_throw_adhoc(tc, "Concurrent modification of UTF-8 input buffer!");
+            utf8_decode_errors(tc, orig_utf8, orig_bytes);
             break;
         }
     }
@@ -254,35 +279,49 @@ MVMString * MVM_string_utf8_decode(MVMThreadContext *tc, const MVMObject *result
     MVM_unicode_normalizer_eof(tc, &norm);
     ready = MVM_unicode_normalizer_available(tc, &norm);
     if (ready) {
-        if (count + ready >= bufsize) {
-            buffer = MVM_realloc(buffer, sizeof(MVMGrapheme32) * (count + ready));
+        while (ready--) {
+            buffer[count++] =  MVM_unicode_normalizer_get_grapheme(tc, &norm);
         }
-        while (ready--)
-            buffer[count++] = MVM_unicode_normalizer_get_grapheme(tc, &norm);
     }
     MVM_unicode_normalizer_cleanup(tc, &norm);
 
-    /* just keep the same buffer as the MVMString's buffer.  Later
-     * we can add heuristics to resize it if we have enough free
-     * memory */
-    if (bufsize - count > 4) {
-        buffer = MVM_realloc(buffer, count * sizeof(MVMGrapheme32));
+    /* If we're lucky, we can fit our string in 8 bits per grapheme. */
+    if (MVM_string_buf32_can_fit_into_8bit(buffer, count)) {
+        MVMGrapheme8 *storage;
+        if (count <= 8) {
+            storage = result->body.storage.in_situ_8;
+            result->body.storage_type    = MVM_STRING_IN_SITU_8;
+        } else {
+            storage = result->body.storage.blob_8 = MVM_malloc(sizeof(MVMGrapheme8) * count);
+            result->body.storage_type    = MVM_STRING_GRAPHEME_8;
+        }
+        MVM_VECTORIZE_LOOP
+        for (ready = 0; ready < count; ready++) {
+            storage[ready] = buffer[ready];
+        }
+        MVM_free(buffer);
+    } else {
+        /* just keep the same buffer as the MVMString's buffer.  Later
+         * we can add heuristics to resize it if we have enough free
+         * memory */
+        if (bufsize - count > 8) {
+            buffer = MVM_realloc(buffer, count * sizeof(MVMGrapheme32));
+        }
+        result->body.storage.blob_32 = buffer;
+        result->body.storage_type    = MVM_STRING_GRAPHEME_32;
     }
-    result->body.storage.blob_32 = buffer;
-    result->body.storage_type    = MVM_STRING_GRAPHEME_32;
     result->body.num_graphs      = count;
 
     return result;
 }
 
-static MVMint32 its_the_bom(const char *utf8) {
-    const MVMuint8 *uns_utf8 = (const MVMuint8 *)utf8;
-    return uns_utf8[0] == 0xEF && uns_utf8[1] == 0xBB && uns_utf8[2] == 0xBF;
+static MVMint32 its_the_bom(const MVMuint8 *utf8) {
+    return utf8[0] == 0xEF && utf8[1] == 0xBB && utf8[2] == 0xBF;
 }
 
 /* Same as MVM_string_utf8_decode, but strips a BOM if it finds one. */
 MVMString * MVM_string_utf8_decode_strip_bom(MVMThreadContext *tc, const MVMObject *result_type, const char *utf8, size_t bytes) {
-    if (bytes >= 3 && its_the_bom(utf8)) {
+    if (bytes >= 3 && its_the_bom((MVMuint8*)utf8)) {
         utf8 += 3;
         bytes -= 3;
     }
@@ -291,38 +330,46 @@ MVMString * MVM_string_utf8_decode_strip_bom(MVMThreadContext *tc, const MVMObje
 
 /* Decodes using a decodestream. Decodes as far as it can with the input
  * buffers, or until a stopper is reached. */
-void MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
-                                  const MVMint32 *stopper_chars,
+MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
+                                  const MVMuint32 *stopper_chars,
                                   MVMDecodeStreamSeparators *seps) {
-    MVMint32 count = 0, total = 0;
+    MVMuint32 count = 0, total = 0;
     MVMint32 state = 0;
     MVMCodepoint codepoint = 0;
-    MVMint32 bufsize;
-    MVMGrapheme32 *buffer;
-    MVMDecodeStreamBytes *cur_bytes;
-    MVMDecodeStreamBytes *last_accept_bytes = ds->bytes_head;
-    MVMint32 last_accept_pos, ready, at_start;
+    MVMCodepoint lag_codepoint = -1;
+    MVMuint32 bufsize;
+    MVMGrapheme32 *buffer           = NULL;
+    MVMDecodeStreamBytes *cur_bytes = NULL;
+    MVMDecodeStreamBytes *last_accept_bytes     = ds->bytes_head;
+    MVMDecodeStreamBytes *lag_last_accept_bytes = NULL;
+    MVMint32 last_accept_pos, lag_last_accept_pos = 0, ready, at_start;
+    MVMuint32 reached_stopper;
+    MVMuint32 can_fast_path;
 
     /* If there's no buffers, we're done. */
     if (!ds->bytes_head)
-        return;
+        return 0;
     last_accept_pos = ds->bytes_head_pos;
 
     /* If we're asked for zero chars, also done. */
     if (stopper_chars && *stopper_chars == 0)
-        return;
+        return 1;
 
-    /* Rough starting-size estimate is number of bytes in the head buffer. */
-    bufsize = ds->bytes_head->length;
+    /* If there's nothing hanging around in the normalization buffer, we can
+     * use the fast path. */
+    can_fast_path = MVM_unicode_normalizer_empty(tc, &(ds->norm));
+
+    bufsize = ds->result_size_guess;
     buffer = MVM_malloc(bufsize * sizeof(MVMGrapheme32));
 
     /* Decode each of the buffers. */
     cur_bytes = ds->bytes_head;
     at_start = ds->abs_byte_pos == 0;
+    reached_stopper = 0;
     while (cur_bytes) {
         /* Process this buffer. */
         MVMint32  pos   = cur_bytes == ds->bytes_head ? ds->bytes_head_pos : 0;
-        char     *bytes = cur_bytes->bytes;
+        MVMuint8 *bytes = cur_bytes->bytes;
         if (at_start) {
             /* We're right at the start of the stream of things to decode. See
              * if we have a BOM, and skip over it if so. */
@@ -335,19 +382,58 @@ void MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
             }
             at_start = 0;
         }
-        while (pos < cur_bytes->length) {
-            switch(decode_utf8_byte(&state, &codepoint, bytes[pos++])) {
-            case UTF8_ACCEPT: {
-                MVMint32 first = 1;
-                MVMGrapheme32 g;
-                last_accept_bytes = cur_bytes;
-                last_accept_pos = pos;
-                ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &(ds->norm), codepoint, &g);
-                while (ready--) {
-                    if (first)
-                        first = 0;
-                    else
-                        g = MVM_unicode_normalizer_get_grapheme(tc, &(ds->norm));
+
+        /* We have both a fast path and a slow path for UTF-8 decoding. The
+         * fast path covers the common case where we have no chars that are
+         * significant to normalization, and so we can skip the normalizer.
+         * This is true of the ASCII and Latin-1 ranges of UTF-8, with the
+         * exception of \r. Note that since the following codepoint may be
+         * the one that causes us to need to compose, we need a lag of 1
+         * codepoint. */
+        if (can_fast_path) {
+            /* Lift the no lag codepoint case out of the hot loop below,
+             * to save on a couple of branches. */
+            MVMCodepoint first_significant = ds->norm.first_significant;
+            while (lag_codepoint == -1 && pos < cur_bytes->length) {
+                switch(MVM_EXPECT(decode_utf8_byte(&state, &codepoint, bytes[pos++]), UTF8_ACCEPT)) {
+                case UTF8_ACCEPT: {
+                    if (codepoint == '\r' || codepoint >= first_significant) {
+                        can_fast_path = 0;
+                        last_accept_bytes = cur_bytes;
+                        last_accept_pos = pos;
+                        goto slow_path;
+                    }
+                    lag_codepoint = codepoint;
+                    lag_last_accept_bytes = cur_bytes;
+                    lag_last_accept_pos = pos;
+                    break;
+                }
+                case UTF8_REJECT: {
+                    MVM_free(buffer);
+                    encoding_error(tc, bytes, pos - 1, -1, -1);
+                    break;
+                }
+                }
+            }
+
+            while (pos < cur_bytes->length) {
+                switch(MVM_EXPECT(decode_utf8_byte(&state, &codepoint, bytes[pos++]), UTF8_ACCEPT)) {
+                case UTF8_ACCEPT: {
+                    /* If we hit something that needs the normalizer, we put
+                     * any lagging codepoint into its buffer and jump to it. */
+                    if (codepoint == '\r' || codepoint >= first_significant) {
+                        MVM_unicode_normalizer_push_codepoints(tc, &(ds->norm),
+                            &lag_codepoint, 1);
+                        lag_codepoint = -1; /* Invalidate, we used it. */
+                        can_fast_path = 0;
+                        last_accept_bytes = cur_bytes;
+                        last_accept_pos = pos;
+                        goto slow_path;
+                    }
+
+                    /* As we have a lagging codepoint, and this one does not
+                     * need normalization, then we know we can spit out the
+                     * lagging one. */
                     if (count == bufsize) {
                         /* Valid character, but we filled the buffer. Attach this
                         * one to the buffers linked list, and continue with a new
@@ -356,18 +442,83 @@ void MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
                         buffer = MVM_malloc(bufsize * sizeof(MVMGrapheme32));
                         count = 0;
                     }
-                    buffer[count++] = g;
+                    buffer[count++] = lag_codepoint;
                     total++;
-                    if (stopper_chars && *stopper_chars == total)
-                        goto done;
-                    if (MVM_string_decode_stream_maybe_sep(tc, seps, g))
+                    if (MVM_string_decode_stream_maybe_sep(tc, seps, lag_codepoint) ||
+                            (stopper_chars && *stopper_chars == total)) {
+                        reached_stopper = 1;
+                        last_accept_bytes = lag_last_accept_bytes;
+                        last_accept_pos = lag_last_accept_pos;
                         goto done;
+                    }
+
+                    /* The current state becomes the lagged state. */
+                    lag_codepoint = codepoint;
+                    lag_last_accept_bytes = cur_bytes;
+                    lag_last_accept_pos = pos;
+                    break;
+                }
+                case UTF8_REJECT: {
+                    MVM_free(buffer);
+                    encoding_error(tc, bytes, pos - 1, -1, -1);
+                    break;
+                }
                 }
-                break;
             }
-            case UTF8_REJECT:
-                MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
-                break;
+
+            /* If we fall out of the loop and have a lagged codepoint, but
+             * no next buffer, then we fall into the slow path to process it
+             * correctly. */
+            if (lag_codepoint != -1 && !cur_bytes->next) {
+                codepoint = lag_codepoint;
+                lag_codepoint = -1;
+                can_fast_path = 0;
+                last_accept_bytes = lag_last_accept_bytes;
+                last_accept_pos = lag_last_accept_pos;
+                goto slow_path;
+            }
+        }
+        else {
+            while (pos < cur_bytes->length) {
+                switch(MVM_EXPECT(decode_utf8_byte(&state, &codepoint, bytes[pos++]), UTF8_ACCEPT)) {
+                case UTF8_ACCEPT: {
+                    MVMGrapheme32 g;
+                    MVMint32 first;
+                    last_accept_bytes = cur_bytes;
+                    last_accept_pos = pos;
+                  slow_path:
+                    first = 1;
+                    ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc,
+                        &(ds->norm), codepoint, &g);
+                    while (ready--) {
+                        if (first)
+                            first = 0;
+                        else
+                            g = MVM_unicode_normalizer_get_grapheme(tc, &(ds->norm));
+                        if (count == bufsize) {
+                            /* Valid character, but we filled the buffer. Attach this
+                            * one to the buffers linked list, and continue with a new
+                            * one. */
+                            MVM_string_decodestream_add_chars(tc, ds, buffer, bufsize);
+                            buffer = MVM_malloc(bufsize * sizeof(MVMGrapheme32));
+                            count = 0;
+                        }
+                        buffer[count++] = g;
+                        total++;
+                        if (MVM_string_decode_stream_maybe_sep(tc, seps, g) ||
+                                (stopper_chars && *stopper_chars == total)) {
+                            reached_stopper = 1;
+                            goto done;
+                        }
+                    }
+                    break;
+                }
+                case UTF8_REJECT: {
+                    MVM_free(buffer);
+                    encoding_error(tc, bytes, pos - 1, -1, -1);
+                    break;
+                }
+                }
             }
         }
         cur_bytes = cur_bytes->next;
@@ -383,25 +534,27 @@ void MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
         MVM_free(buffer);
     }
     MVM_string_decodestream_discard_to(tc, ds, last_accept_bytes, last_accept_pos);
+
+    return reached_stopper;
 }
 
 /* Encodes the specified string to UTF-8. */
 char * MVM_string_utf8_encode_substr(MVMThreadContext *tc,
         MVMString *str, MVMuint64 *output_size, MVMint64 start, MVMint64 length,
         MVMString *replacement, MVMint32 translate_newlines) {
-    MVMuint8        *result;
+    MVMuint8        *result = NULL;
     size_t           result_pos, result_limit;
     MVMCodepointIter ci;
-    MVMStringIndex   strgraphs = MVM_string_graphs(tc, str);
+    MVMStringIndex   strgraphs  = MVM_string_graphs(tc, str);
     MVMuint8        *repl_bytes = NULL;
     MVMuint64        repl_length;
 
     if (start < 0 || start > strgraphs)
-        MVM_exception_throw_adhoc(tc, "start out of range");
+        MVM_exception_throw_adhoc(tc, "start (%"PRId64") out of range (0..%"PRIu32")", start, strgraphs);
     if (length == -1)
         length = strgraphs;
     if (length < 0 || start + length > strgraphs)
-        MVM_exception_throw_adhoc(tc, "length out of range");
+        MVM_exception_throw_adhoc(tc, "length (%"PRId64") out of range (0..%"PRIu32")", length, strgraphs);
 
     if (replacement)
         repl_bytes = (MVMuint8 *) MVM_string_utf8_encode_substr(tc,
@@ -414,7 +567,7 @@ char * MVM_string_utf8_encode_substr(MVMThreadContext *tc,
     result_pos   = 0;
 
     /* Iterate the codepoints and encode them. */
-    MVM_string_ci_init(tc, &ci, str, translate_newlines);
+    MVM_string_ci_init(tc, &ci, str, translate_newlines, 0);
     while (MVM_string_ci_has_more(tc, &ci)) {
         MVMint32 bytes;
         MVMCodepoint cp = MVM_string_ci_get_codepoint(tc, &ci);
@@ -436,9 +589,7 @@ char * MVM_string_utf8_encode_substr(MVMThreadContext *tc,
         else {
             MVM_free(result);
             MVM_free(repl_bytes);
-            MVM_exception_throw_adhoc(tc,
-                "Error encoding UTF-8 string: could not encode codepoint %d",
-                cp);
+            MVM_string_utf8_throw_encoding_exception(tc, cp);
         }
     }
 
@@ -458,13 +609,78 @@ char * MVM_string_utf8_encode(MVMThreadContext *tc, MVMString *str, MVMuint64 *o
 /* Encodes the specified string to a UTF-8 C string. */
 char * MVM_string_utf8_encode_C_string(MVMThreadContext *tc, MVMString *str) {
     MVMuint64 output_size;
-    char * result;
     char * utf8_string = MVM_string_utf8_encode(tc, str, &output_size, 0);
     /* this is almost always called from error-handling code. Don't care if it
      * contains embedded NULs. XXX TODO: Make sure all uses of this free what it returns */
-    result = MVM_malloc(output_size + 1);
-    memcpy(result, utf8_string, output_size);
-    MVM_free(utf8_string);
-    result[output_size] = (char)0;
-    return result;
+    utf8_string = MVM_realloc(utf8_string, output_size + 1);
+    utf8_string[output_size] = (char)0;
+    return utf8_string;
 }
+
+/* Encodes the specified string to a UTF-8 C string. */
+char * MVM_string_utf8_encode_C_string_malloc(MVMThreadContext *tc, MVMString *str) {
+    MVMint64         length = MVM_string_graphs(tc, str);
+    /* Guesstimate that we'll be within 2 bytes for most chars most of the
+     * time, and give ourselves 4 bytes breathing space, plus 1 for the NUL. */
+    size_t           result_limit = 2 * length;
+    MVMuint8        *result = malloc(result_limit + 4 + 1);
+    size_t           result_pos = 0;
+
+    /* Iterate the codepoints and encode them. */
+    MVMCodepointIter ci;
+    MVM_string_ci_init(tc, &ci, str, 0, 0);
+    while (MVM_string_ci_has_more(tc, &ci)) {
+        MVMCodepoint cp = MVM_string_ci_get_codepoint(tc, &ci);
+        if (result_pos >= result_limit) {
+            result_limit *= 2;
+            result = realloc(result, result_limit + 4 + 1);
+        }
+        MVMint32 bytes = utf8_encode(result + result_pos, cp);
+        if (bytes)
+            result_pos += bytes;
+        else {
+            free(result);
+            MVM_string_utf8_throw_encoding_exception(tc, cp);
+        }
+    }
+
+    result[result_pos] = 0;
+    return (char *)result;
+}
+
+/* Encodes the specified string to a UTF-8 C string if it is not NULL. */
+char * MVM_string_utf8_maybe_encode_C_string(MVMThreadContext *tc, MVMString *str) {
+    return str ? MVM_string_utf8_encode_C_string(tc, str) : NULL;
+}
+
+void MVM_string_utf8_throw_encoding_exception (MVMThreadContext *tc, MVMCodepoint cp) {
+    const char *gencat = MVM_unicode_codepoint_get_property_cstr(tc, cp, MVM_UNICODE_PROPERTY_GENERAL_CATEGORY);
+    if(cp > 0x10FFFF) {
+        MVM_exception_throw_adhoc(tc,
+            "Error encoding UTF-8 string: could not encode codepoint %d (0x%X), codepoint out of bounds. Cannot encode higher than %d (0x%X)",
+            cp, cp, 0x10FFFF, 0x10FFFF);
+    }
+    else if (strcmp("Cs", gencat) == 0) {
+        MVM_exception_throw_adhoc(tc,
+            "Error encoding UTF-8 string: could not encode Unicode Surrogate codepoint %d (0x%X)",
+            cp, cp);
+    }
+    else {
+        MVM_exception_throw_adhoc(tc,
+            "Error encoding UTF-8 string: could not encode codepoint %d (0x%X)",
+            cp, cp);
+    }
+}
+
+#ifdef DEBUG_HELPERS
+void MVM_dump_string(MVMThreadContext *tc, MVMString *s) {
+    char *encoded = MVM_string_utf8_maybe_encode_C_string(tc, s);
+    if (encoded) {
+        fprintf(stderr, "%s\n", encoded);
+        MVM_free(encoded);
+    }
+    else {
+        fprintf(stderr, "<null>\n");
+    }
+}
+#endif
